@@ -4,38 +4,29 @@ import os
 import socket
 import sys
 import uuid
-import warnings
-
-from future.standard_library import install_aliases
-
-from pyngrok import process, conf
-from pyngrok.exception import PyngrokNgrokHTTPError, PyngrokNgrokURLError, PyngrokSecurityError, PyngrokError
-from pyngrok.installer import install_ngrok, install_default_config
-
-install_aliases()
-
+from http import HTTPStatus
+from urllib.error import HTTPError, URLError
 from urllib.parse import urlencode
-from urllib.request import urlopen, Request, HTTPError, URLError
+from urllib.request import urlopen, Request
 
-try:
-    from http import HTTPStatus as StatusCodes
-except ImportError:  # pragma: no cover
-    try:
-        from http import client as StatusCodes
-    except ImportError:
-        import httplib as StatusCodes
+from pyngrok import process, conf, installer
+from pyngrok.exception import PyngrokNgrokHTTPError, PyngrokNgrokURLError, PyngrokSecurityError, PyngrokError
 
 __author__ = "Alex Laird"
 __copyright__ = "Copyright 2020, Alex Laird"
-__version__ = "4.2.2"
+__version__ = "5.0.0"
 
 logger = logging.getLogger(__name__)
 
+_current_tunnels = {}
+
 
 class NgrokTunnel:
     """
     An object containing information about a ``ngrok`` tunnel.
 
+    :var data: The original tunnel data.
+    :vartype data: dict
     :var name: The name of the tunnel.
     :vartype name: str
     :var proto: A valid `tunnel protocol <https://ngrok.com/docs#tunnel-definitions>`_.
@@ -48,17 +39,14 @@ class NgrokTunnel:
     :vartype config: dict
     :var metrics: Metrics for `the tunnel <https://ngrok.com/docs#list-tunnels>`_.
     :vartype metrics: dict
-    :var pyngrok_config: The ``pyngrok`` configuration to use with ``ngrok``.
+    :var pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok``.
     :vartype pyngrok_config: PyngrokConfig
     :var api_url: The API URL for the ``ngrok`` web interface.
     :vartype api_url: str
     """
 
-    def __init__(self, data=None, pyngrok_config=None, api_url=None):
-        if data is None:
-            data = {}
-        if pyngrok_config is None:
-            pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+    def __init__(self, data, pyngrok_config, api_url):
+        self.data = data
 
         self.name = data.get("name")
         self.proto = data.get("proto")
@@ -66,6 +54,7 @@ class NgrokTunnel:
         self.public_url = data.get("public_url")
         self.config = data.get("config", {})
         self.metrics = data.get("metrics", {})
+
         self.pyngrok_config = pyngrok_config
         self.api_url = api_url
 
@@ -79,33 +68,49 @@ class NgrokTunnel:
 
     def refresh_metrics(self):
         """
-        Refresh the metrics from the tunnel.
+        Get the latest metrics for the tunnel and update the ``metrics`` variable.
         """
-        if self.api_url is None:
-            raise PyngrokError("\"api_url\" was not initialized with this NgrokTunnel, so this method cannot be used.")
+        logger.info("Refreshing metrics for tunnel: {}".format(self.public_url))
 
-        data = api_request("{}{}".format(self.api_url, self.uri), method="GET", data=None,
+        data = api_request("{}{}".format(self.api_url, self.uri), method="GET",
                            timeout=self.pyngrok_config.request_timeout)
 
         if "metrics" not in data:
             raise PyngrokError("The ngrok API did not return \"metrics\" in the response")
 
-        self.metrics = data["metrics"]
+        self.data["metrics"] = data["metrics"]
+        self.metrics = self.data["metrics"]
 
 
-def ensure_ngrok_installed(ngrok_path):
+def install_ngrok(pyngrok_config=None):
     """
-    Ensure ``ngrok`` is installed at the given path, downloading and installing the binary for
-    the current system if not.
+    Download, install, and initialize ``ngrok`` for the given config. If ``ngrok`` and its default
+    config is already installed, calling this method will do nothing.
 
-    :param ngrok_path: The path to the ``ngrok`` binary.
-    :type ngrok_path: str
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
+    :type pyngrok_config: PyngrokConfig, optional
     """
-    if not os.path.exists(ngrok_path):
-        install_ngrok(ngrok_path)
+    if pyngrok_config is None:
+        pyngrok_config = conf.get_default()
 
-    if not os.path.exists(conf.DEFAULT_NGROK_CONFIG_PATH):
-        install_default_config(conf.DEFAULT_NGROK_CONFIG_PATH)
+    if not os.path.exists(pyngrok_config.ngrok_path):
+        installer.install_ngrok(pyngrok_config.ngrok_path)
+
+    # If no config_path is set, ngrok will use its default path
+    if pyngrok_config.config_path is not None:
+        config_path = pyngrok_config.config_path
+    else:
+        config_path = conf.DEFAULT_NGROK_CONFIG_PATH
+
+    # Install the config to the requested path
+    if not os.path.exists(config_path):
+        installer.install_default_config(config_path)
+
+    # Install the default config, even if we don't need it this time, if it doesn't already exist
+    if conf.DEFAULT_NGROK_CONFIG_PATH != config_path and \
+            not os.path.exists(conf.DEFAULT_NGROK_CONFIG_PATH):
+        installer.install_default_config(conf.DEFAULT_NGROK_CONFIG_PATH)
 
 
 def set_auth_token(token, pyngrok_config=None):
@@ -118,22 +123,21 @@ def set_auth_token(token, pyngrok_config=None):
 
     :param token: The auth token to set.
     :type token: str
-    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
-        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
-        `as shown here <index.html#config-file>`_).
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
     :type pyngrok_config: PyngrokConfig, optional
     """
     if pyngrok_config is None:
-        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+        pyngrok_config = conf.get_default()
 
-    ensure_ngrok_installed(pyngrok_config.ngrok_path)
+    install_ngrok(pyngrok_config)
 
     process.set_auth_token(pyngrok_config, token)
 
 
 def get_ngrok_process(pyngrok_config=None):
     """
-    Retrieve the current ``ngrok`` process for the given config's ``ngrok_path``.
+    Get the current ``ngrok`` process for the given config's ``ngrok_path``.
 
     If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
     will first download and install ``ngrok``.
@@ -141,25 +145,32 @@ def get_ngrok_process(pyngrok_config=None):
     If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
-    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
-        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
-        `as shown here <index.html#config-file>`_).
+    Use :func:`~pyngrok.process.is_process_running` to check if a process is running without also implicitly
+    installing and starting it.
+
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
     :type pyngrok_config: PyngrokConfig, optional
     :return: The ``ngrok`` process.
     :rtype: NgrokProcess
     """
     if pyngrok_config is None:
-        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+        pyngrok_config = conf.get_default()
 
-    ensure_ngrok_installed(pyngrok_config.ngrok_path)
+    install_ngrok(pyngrok_config)
 
     return process.get_process(pyngrok_config)
 
 
-def connect(port="80", proto="http", name=None, pyngrok_config=None, return_ngrok_tunnel=False, **options):
+def connect(addr=None, proto=None, name=None, pyngrok_config=None, **options):
     """
-    Establish a new ``ngrok`` tunnel for the given protocol to the given port, returning the connected
-    public URL that tunnels to the local port.
+    Establish a new ``ngrok`` tunnel for the given protocol to the given port, returning a object representing
+    the connected tunnel.
+
+    If a `tunnel definition in ngrok's config file <https://ngrok.com/docs#tunnel-definitions>`_ matches the given
+    ``name``, it will be loaded and used to start the tunnel. When ``name`` is ``None`` and a "pyngrok-default" tunnel
+    definition exists in ``ngrok``'s config, it will be loaded and use. Any ``kwargs`` passed as ``options`` will
+    override properties from the loaded tunnel definition.
 
     If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
     will first download and install ``ngrok``.
@@ -167,107 +178,127 @@ def connect(port="80", proto="http", name=None, pyngrok_config=None, return_ngro
     If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
-    :param port: The local port to which the tunnel will forward traffic, defaults to "80". Can also be
-        a `local directory or network address <https://ngrok.com/docs#http-file-urls>`_.
-    :type port: str, optional
+    :param addr: The local port to which the tunnel will forward traffic, or a
+        `local directory or network address <https://ngrok.com/docs#http-file-urls>`_, defaults to "80".
+    :type addr: str, optional
     :param proto: The protocol to tunnel, defaults to "http".
     :type proto: str, optional
-    :param name: A friendly name for the tunnel.
+    :param name: A friendly name for the tunnel, or the name of a `ngrok tunnel definition <https://ngrok.com/docs#tunnel-definitions>`_
+        to be used.
     :type name: str, optional
-    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
-        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
-        `as shown here <index.html#config-file>`_).
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
     :type pyngrok_config: PyngrokConfig, optional
-    :param return_ngrok_tunnel: ``True`` if a :class:`~pyngrok.ngrok.NgrokTunnel` should be returned, ``False`` for ``str``.
-    :type return_ngrok_tunnel: bool, optional
     :param options: Remaining ``kwargs`` are passed as `configuration for the ngrok
         tunnel <https://ngrok.com/docs#tunnel-definitions>`_.
     :type options: dict, optional
-    :return: The connected public URL if ``return_ngrok_tunnel`` is ``False``, otherwise a :class:`~pyngrok.ngrok.NgrokTunnel`.
-    :rtype: str
+    :return: The created ``ngrok`` tunnel.
+    :rtype: NgrokTunnel
     """
-    if options is None:
-        options = {}
     if pyngrok_config is None:
-        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+        pyngrok_config = conf.get_default()
 
-    if "options" in options:
-        warnings.warn(
-            "Support for passing \"options\" as a dict is deprecated and will be removed in 5.0.0, unpack the dict as kwargs",
-            DeprecationWarning, 2)
-        options.update(options.pop("options", {}))
+    if pyngrok_config.config_path is not None:
+        config_path = pyngrok_config.config_path
+    else:
+        config_path = conf.DEFAULT_NGROK_CONFIG_PATH
+
+    if os.path.exists(config_path):
+        config = installer.get_ngrok_config(config_path)
+    else:
+        config = {}
+
+    # If a "pyngrok-default" tunnel definition exists in the ngrok config, use that
+    tunnel_definitions = config.get("tunnels", {})
+    if not name and "pyngrok-default" in tunnel_definitions:
+        name = "pyngrok-default"
+
+    # Use a tunnel definition for the given name, if it exists
+    if name and name in tunnel_definitions:
+        tunnel_definition = tunnel_definitions[name]
+
+        addr = tunnel_definition.get("addr") if not addr else addr
+        proto = tunnel_definition.get("proto") if not proto else proto
+        # Use the tunnel definition as the base, but override with any passed in options
+        tunnel_definition.update(options)
+        options = tunnel_definition
+
+    addr = str(addr) if addr else "80"
+    if not proto:
+        proto = "http"
 
-    port = str(port)
     if not name:
-        if not port.startswith("file://"):
-            name = "{}-{}-{}".format(proto, port, uuid.uuid4())
+        if not addr.startswith("file://"):
+            name = "{}-{}-{}".format(proto, addr, uuid.uuid4())
         else:
             name = "{}-file-{}".format(proto, uuid.uuid4())
 
+    logger.info("Opening tunnel named: {}".format(name))
+
     config = {
         "name": name,
-        "addr": port,
+        "addr": addr,
         "proto": proto
     }
     options.update(config)
 
     api_url = get_ngrok_process(pyngrok_config).api_url
 
-    logger.debug("Connecting tunnel with options: {}".format(options))
+    logger.debug("Creating tunnel with options: {}".format(options))
 
     tunnel = NgrokTunnel(api_request("{}/api/tunnels".format(api_url), method="POST", data=options,
                                      timeout=pyngrok_config.request_timeout),
-                         pyngrok_config=pyngrok_config, api_url=api_url)
+                         pyngrok_config, api_url)
 
     if proto == "http" and not options.get("bind_tls", False):
         tunnel = NgrokTunnel(api_request("{}{}%20%28http%29".format(api_url, tunnel.uri), method="GET",
                                          timeout=pyngrok_config.request_timeout),
                              pyngrok_config, api_url)
 
-    if return_ngrok_tunnel:
-        return tunnel
-    else:
-        warnings.warn("Support for \"return_ngrok_tunnel\" as \"False\" is deprecated and will be removed in 5.0.0, "
-                      "when this method will return a NgrokTunnel instead of a str", DeprecationWarning, 2)
-        return tunnel.public_url
+    _current_tunnels[tunnel.public_url] = tunnel
+
+    return tunnel
 
 
 def disconnect(public_url, pyngrok_config=None):
     """
-    Disconnect the ``ngrok`` tunnel for the given URL.
-
-    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
-    will first download and install ``ngrok``.
-
-    If ``ngrok`` is not running, calling this method will first start a process with
-    :class:`~pyngrok.conf.PyngrokConfig`.
+    Disconnect the ``ngrok`` tunnel for the given URL, if open.
 
     :param public_url: The public URL of the tunnel to disconnect.
     :type public_url: str
-    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
-        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
-        `as shown here <index.html#config-file>`_).
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
     :type pyngrok_config: PyngrokConfig, optional
     """
     if pyngrok_config is None:
-        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+        pyngrok_config = conf.get_default()
+
+    # If ngrok is not running, there are no tunnels to disconnect
+    if not process.is_process_running(pyngrok_config.ngrok_path):
+        return
 
     api_url = get_ngrok_process(pyngrok_config).api_url
 
-    tunnels = get_tunnels(pyngrok_config)
-    for tunnel in tunnels:
-        if tunnel.public_url == public_url:
-            logger.debug("Disconnecting tunnel: {}".format(tunnel.public_url))
+    if public_url not in _current_tunnels:
+        get_tunnels(pyngrok_config)
 
-            api_request("{}{}".format(api_url, tunnel.uri), method="DELETE",
-                        timeout=pyngrok_config.request_timeout)
+        # One more check, if the given URL is still not in the list of tunnels, it is not active
+        if public_url not in _current_tunnels:
+            return
 
-            break
+    tunnel = _current_tunnels[public_url]
+
+    logger.info("Disconnecting tunnel: {}".format(tunnel.public_url))
+
+    api_request("{}{}".format(api_url, tunnel.uri), method="DELETE",
+                timeout=pyngrok_config.request_timeout)
+
+    _current_tunnels.pop(public_url, None)
 
 
 def get_tunnels(pyngrok_config=None):
     """
-    Retrieve a list of active ``ngrok`` tunnels for the given config's ``ngrok_path``.
+    Get a list of active ``ngrok`` tunnels for the given config's ``ngrok_path``.
 
     If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
     will first download and install ``ngrok``.
@@ -275,24 +306,24 @@ def get_tunnels(pyngrok_config=None):
     If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
-    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
-        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
-        `as shown here <index.html#config-file>`_).
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
     :type pyngrok_config: PyngrokConfig, optional
     :return: The active ``ngrok`` tunnels.
     :rtype: list[NgrokTunnel]
     """
     if pyngrok_config is None:
-        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+        pyngrok_config = conf.get_default()
 
     api_url = get_ngrok_process(pyngrok_config).api_url
 
-    tunnels = []
-    for tunnel in api_request("{}/api/tunnels".format(api_url), method="GET", data=None,
+    _current_tunnels.clear()
+    for tunnel in api_request("{}/api/tunnels".format(api_url), method="GET",
                               timeout=pyngrok_config.request_timeout)["tunnels"]:
-        tunnels.append(NgrokTunnel(tunnel, pyngrok_config=pyngrok_config, api_url=api_url))
+        ngrok_tunnel = NgrokTunnel(tunnel, pyngrok_config, api_url)
+        _current_tunnels[ngrok_tunnel.public_url] = ngrok_tunnel
 
-    return tunnels
+    return list(_current_tunnels.values())
 
 
 def kill(pyngrok_config=None):
@@ -300,20 +331,55 @@ def kill(pyngrok_config=None):
     Terminate the ``ngrok`` processes, if running, for the given config's ``ngrok_path``. This method will not
     block, it will just issue a kill request.
 
-    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
-        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
-        `as shown here <index.html#config-file>`_).
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
     :type pyngrok_config: PyngrokConfig, optional
     """
     if pyngrok_config is None:
-        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
+        pyngrok_config = conf.get_default()
 
     process.kill_process(pyngrok_config.ngrok_path)
 
+    _current_tunnels.clear()
+
+
+def get_version(pyngrok_config=None):
+    """
+    Get a tuple with the ``ngrok`` and ``pyngrok`` versions.
+
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
+    :type pyngrok_config: PyngrokConfig, optional
+    :return: A tuple of ``(ngrok_version, pyngrok_version)``.
+    :rtype: tuple
+    """
+    if pyngrok_config is None:
+        pyngrok_config = conf.get_default()
+
+    ngrok_version = process.capture_run_process(pyngrok_config.ngrok_path, ["--version"]).split("version ")[1]
+
+    return ngrok_version, __version__
+
+
+def update(pyngrok_config=None):
+    """
+    Update ``ngrok`` for the given config's ``ngrok_path``, if an update is available.
+
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
+    :type pyngrok_config: PyngrokConfig, optional
+    :return: The result from the ``ngrok`` update.
+    :rtype: str
+    """
+    if pyngrok_config is None:
+        pyngrok_config = conf.get_default()
+
+    return process.capture_run_process(pyngrok_config.ngrok_path, ["update"])
+
 
 def api_request(url, method="GET", data=None, params=None, timeout=4):
     """
-    Invoke an API request to the given URL, returning JSON data from the response as a dict.
+    Invoke an API request to the given URL, returning JSON data from the response.
 
     One use for this method is making requests to ``ngrok`` tunnels:
 
@@ -369,13 +435,12 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
         response_data = response.read().decode("utf-8")
 
         status_code = response.getcode()
-        logger.debug("Response status code: {}".format(status_code))
-        logger.debug("Response: {}".format(response_data.strip()))
+        logger.debug("Response {}: {}".format(status_code, response_data.strip()))
 
         if str(status_code)[0] != "2":
             raise PyngrokNgrokHTTPError("ngrok client API returned {}: {}".format(status_code, response_data), url,
                                         status_code, None, request.headers, response_data)
-        elif status_code == StatusCodes.NO_CONTENT:
+        elif status_code == HTTPStatus.NO_CONTENT:
             return None
 
         return json.loads(response_data)
@@ -385,8 +450,7 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
         response_data = e.read().decode("utf-8")
 
         status_code = e.getcode()
-        logger.debug("Response status code: {}".format(status_code))
-        logger.debug("Response: {}".format(response_data.strip()))
+        logger.debug("Response {}: {}".format(status_code, response_data.strip()))
 
         raise PyngrokNgrokHTTPError("ngrok client exception, API returned {}: {}".format(status_code, response_data),
                                     e.url,
@@ -395,9 +459,9 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
         raise PyngrokNgrokURLError("ngrok client exception, URLError: {}".format(e.reason), e.reason)
 
 
-def run(args=None):
+def run(args=None, pyngrok_config=None):
     """
-    Ensure ``ngrok`` is installed in the default location, then call :func:`~pyngrok.process.run_process`.
+    Ensure ``ngrok`` is installed at the default path, then call :func:`~pyngrok.process.run_process`.
 
     This method is meant for interacting with ``ngrok`` from the command line and is not necessarily
     compatible with non-blocking API methods. For that, use :mod:`~pyngrok.ngrok`'s interface methods (like
@@ -405,13 +469,18 @@ def run(args=None):
 
     :param args: Arguments to be passed to the ``ngrok`` process.
     :type args: list[str], optional
+    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        overriding :func:`~pyngrok.conf.get_default()`.
+    :type pyngrok_config: PyngrokConfig, optional
     """
     if args is None:
         args = []
+    if pyngrok_config is None:
+        pyngrok_config = conf.get_default()
 
-    ensure_ngrok_installed(conf.DEFAULT_NGROK_PATH)
+    install_ngrok(pyngrok_config)
 
-    process.run_process(conf.DEFAULT_NGROK_PATH, args)
+    process.run_process(pyngrok_config.ngrok_path, args)
 
 
 def main():
